/*
 * C6Run.c
 */

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 */

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/************************************************************************
*  Defines GPP(ARM)-side main entry point that process the command line *
*  and start C I/O handling code (loading DSP and serving DSP C I/O)    *
************************************************************************/


/*  ----------------------------------- Standard Headers              */
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <time.h>

/*  ----------------------------------- Linux Posix Headers           */
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>

/*  ----------------------------------- CMEM API                      */
#include <ti/sdo/linuxutils/cmem/include/cmem.h>

/*  ----------------------------------- DSP/BIOS Link                 */
#include <dsplink.h>
#include <errbase.h>

/*  ----------------------------------- Application Headers           */
#include "debug_common.h"
#include "C6Run_ipc.h"
#include "C6Run_cio.h"
#include "control_ipc.h"

/*  ----------------------------------- This module's header          */
#include "C6Run.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/

// External references to dsp image data and size
extern const unsigned char  _dsp_program_data_ [];
extern const unsigned int   _dsp_program_data_size_;


/************************************************************
* Local Macro Declarations                                  *
************************************************************/


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/

// Signal handling routines
static void LOCAL_installSignalHandlers(int numSignals, int *signalList);
static void LOCAL_sigHandler (int sigNum);


/************************************************************
* Local Variable Definitions                                *
************************************************************/

// List of OS signals to handle
static int signalList[8] =
{
  SIGHUP,
  SIGINT,
  SIGILL,
  SIGABRT,
  SIGFPE,
  SIGSEGV,
  SIGALRM,
  SIGTERM
} ;


/************************************************************
* Global Variable Definitions                               *
************************************************************/

volatile bool C6RUN_initialized     = FALSE;
volatile C6RUN_Type C6RUN_type      = C6RUN_TYPE_NONE;

volatile int C6RUN_retVal           = 0;


/************************************************************
* Global Function Definitions                               *
************************************************************/

int C6RUN_appInit(int argc, char **argv)
{
  return C6RUN_init(argc, argv, C6RUN_TYPE_APP);
}

int C6RUN_libInit( void )
{
  return C6RUN_init(0, NULL, C6RUN_TYPE_LIB);
}

int C6RUN_init ( int argc, char **argv, C6RUN_Type progType )
{ 
  DSP_STATUS status = DSP_SOK;

  if (C6RUN_initialized)
  {
    return C6RUN_SUCCESS;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_init()\n");
  
  if ((C6RUN_TYPE_LIB != progType) && (C6RUN_TYPE_APP!= progType))
  {
    return C6RUN_FAIL;
  }

  // Save the waitForReady state
  C6RUN_type = progType;
    
  // Get DSP running, set up MSGQ's, etc.
  status = C6RUN_IPC_create((uint8_t *)_dsp_program_data_, _dsp_program_data_size_, DSP0, argc, argv);
  if (!DSP_SUCCEEDED(status))
  {
    printf("C6RUN_IPC_create() failed!\n");
    exit(1);
  }
  VERBOSE_PRINT0("C6RUN_IPC_create() passed!\n");

  // Start the C6RUN CIO server task
  if (pthread_create(&C6RUN_CIO_thread, NULL, C6RUN_CIO_server, NULL) != 0)
  {
    printf("Aborting...pthread_create() for CIO server failed!\n");
    exit(1);
  }
  VERBOSE_PRINT0("pthread_create() passed!\n");

  // Install the signal handlers we need to corretcly kill this program
  LOCAL_installSignalHandlers(sizeof(signalList)/sizeof(signalList[0]), signalList);
  
  // Send signal to DSP to start
  if (C6RUN_TYPE_LIB == C6RUN_type)
  {
    CONTROL_IPC_send(CONTROL_DSP_STARTLIB);
  }
  else if (C6RUN_TYPE_APP == C6RUN_type)
  {
    CONTROL_IPC_send(CONTROL_DSP_STARTAPP);
  }

  if (C6RUN_TYPE_LIB == C6RUN_type)
  {
    // Pend on receipt of the DSP_READY message for C6RunLib projects
    if (CONTROL_DSP_READY != CONTROL_IPC_recv())
    {
      printf("CONTROL_DSP_READY not recv'd\n");
      pthread_cancel(C6RUN_CIO_thread);
      exit(1);
    }
    VERBOSE_PRINT0("CONTROL_DSP_READY recv'd!\n");
  }
  
  // Register exit routine for calling upon program termination
  atexit(C6RUN_exit);
  
  C6RUN_initialized = TRUE;

  VERBOSE_PRINT0("Leaving C6RUN_init().\n");
  
  return C6RUN_SUCCESS;
}

void C6RUN_exit ( void )
{
  // Exit now if init has not yet been done
  if (!C6RUN_initialized)
  {
    return;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_exit().\n");
  
  // For C6RunLib, send a message to tell the DSP to stop.
  // Otherwise we just wait for the DSP to tell us it is finished.
  if (C6RUN_TYPE_LIB == C6RUN_type)
  {
    CONTROL_IPC_send(CONTROL_DSP_STOP);
  }

  // Pend on receipt of the DSP_FINISHED message
  if (CONTROL_DSP_FINISHED != CONTROL_IPC_recv())
  { 
    // Report if the wrong message ID was received  
    printf("CONTROL_DSP_FINISHED not received.\n");
  }
  else
  {
    VERBOSE_PRINT0("CONTROL_DSP_FINISHED was received.\n");
  }

  // Get the return value if this is C6RunApp
  if (C6RUN_TYPE_APP == C6RUN_type)
  {
    volatile int *p;
  
    // Get return value from the DSP processor
    p = (volatile int *) CONTROL_msg->buffer;
 
    // Message buffer only contains an int32_t return value from main executing on DSP
    C6RUN_retVal = *p;
    
    VERBOSE_PRINT1("DSP return value = %d.\n",C6RUN_retVal);
  }
  
  // Shutdown the C6RUN_CIO_server
  pthread_cancel(C6RUN_CIO_thread);
  
  VERBOSE_PRINT0("Completed shutdown of CIO thread.\n");
  
  // Gracefully shut everything else down (IPC, DSP, etc.)
  C6RUN_IPC_delete(DSP0);

  // Reset the initialized variable
  C6RUN_initialized = false;

  VERBOSE_PRINT0("Leaving C6RUN_exit().\n");
}


/***********************************************************
* Local Function Definitions                               *
***********************************************************/

static void LOCAL_installSignalHandlers(int numSignals, int *signalList)
{
  int i;
  struct sigaction sigInstaller ;
  sigset_t block_mask ;

  sigemptyset( &block_mask );
  sigfillset ( &block_mask );
  sigInstaller.sa_flags   = (SA_RESTART);
  sigInstaller.sa_mask    = block_mask;
  sigInstaller.sa_handler = &LOCAL_sigHandler;

  // Actually register the handler for the different signals
  for (i = 0 ; i < numSignals ; i++)
  {
    sigaction (signalList[i], &sigInstaller, NULL) ;
  }
}

static void LOCAL_sigHandler (int sigNum)
{
  // Output signal handling message
  VERBOSE_PRINT1("Signal %d received! Aborting...",sigNum);
  
  /* It is necessary to call the cleanup function
   * from signal handler context to ensure that the
   * cleanup is not interrupted by another signal.
   * Just exit (0) is not enough  as the at exit handler
   * can be pre-empted by a signal.
   */
  C6RUN_exit ();

  /*  Force hard stop on the process  */
  exit (0);

  return;
}


/***********************************************************
* End file                                                 *
***********************************************************/
